/**
 * 
 * PAUL HUSSMAN
 * 260224172
 * COMP 409 ASSIGNMENT 2 QUESTION 3
 * 
 * 
 * 
 * RESULTS FOR RUNNING THIS CODE:
 * 
 * 		gird dimensions: 30x30
 * 		M = 10
 * 		
 * 				speed	averageMoves	
 * 				=====	============
 * 				1		241027
 * 				2		295229
 * 				3		492078
 * 				.		.
 * 				.		.
 * 				6		602452
 * 				.		.
 * 				8		766269
 * 				.		.
 * 				10		1045797
 * 				.		.
 * 				.		.
 * 				.		.
 * 				15		1114757
 * 
 * 		When objects move at higher speeds, there are less conflicts overall and therefore they get to move more overall.
 * 			It seems that at higher speeds, the diameter of influence between two given objects will be smaller than the
 * 			average distance that objects can reach per unit time, and therefore less conflicts will occur. 
 * 
 * 		This code doesn't deadlock because resources are acquired in order (with the for loop in the run method).
 * 			There will always be someone able to move because the object with the lowest time will never have a conflict, 
 * 			and once he moves, then others can move
 */

import java.util.Random;

public class Point implements Runnable {
	private final int MAX_INIT_TIME = 10000;
	private final int END_OF_TIME;
	private final int X_SIZE;
	private final int Y_SIZE;
	private final int SPEED;
	private final int GRANULARITY = 10;
	private int moveCount = 0;		// keeps track of how many moves this point has made
	private volatile int time;	// current time for this point
	private double xCoordinate;		// current x-position for this point
	private double yCoordinate;		// current y-position for this point
	private Random generator = new Random();
	
	// construction initializes point properties to appropriate random values
	public Point(int totalTime, int xLength, int yLength, int mySpeed) {
		END_OF_TIME = totalTime;
		X_SIZE = xLength;
		Y_SIZE = yLength;
		SPEED = mySpeed;
		time = generator.nextInt(MAX_INIT_TIME+1);
		xCoordinate = X_SIZE*generator.nextDouble();
		yCoordinate = Y_SIZE*generator.nextDouble();
		
	}

	
	public void run(){

		while (time < END_OF_TIME){

			for (int i = 0; i < Torus.pointList.length; i++){
				// checks for time conflict, then distance conflict and both occur calls wait()
				checkConflict(Torus.pointList[i]);
			}
			
			// get a random direction
			int xdir = getRandomDirection(1);
			int ydir = getRandomDirection(xdir);
			
			// move (hypoteneuse)*SPEED pixels per this 10ms (Objects move up, down, left, right or at 45 degree angles by length of radius)
			for (int i = 0; i < (int)getHypoteneuse(xdir, ydir)*SPEED; i++){
				move(xdir, ydir);
			}
			// only increment time if it moves
			incrementTime();
			System.out.println("time ++");
			
			// if this point moves, notify others because they might be able to move now
			synchronized(this){
				notifyAll();
			}
			
			try{Thread.sleep(GRANULARITY);}catch(InterruptedException ie){}
		}

	}
	
	// returns number of moves this point has made
	// 	called only after all threads are joined
	public int getMoveCount(){
		return moveCount;
	}
	// returns this point's current time
	public int getTime(){
		return time;
	}
	// increments this point's current time
	public void incrementTime(){
		time++;
	}
	// moves this point in a random direction
	public synchronized void move(int xdir, int ydir){
		// moves in a random direction ensuring that it actually moves somewhere.
		xCoordinate = (xCoordinate + xdir)%X_SIZE;
		if (xCoordinate < 0) xCoordinate = X_SIZE-xCoordinate;
		yCoordinate = (yCoordinate + ydir)%Y_SIZE;
		if (yCoordinate < 0) yCoordinate = Y_SIZE-yCoordinate;
		moveCount++;
		
	}

	
	// returns true if there is no time conflict between this point and a target point (otherwise false)
	public  boolean isConflict(Point target){
		if(time >= target.getTime()){
			return false;
		}
		if (time - target.getTime() >= getDistance( target.getX(), target.getY() ) ){
			return true;
		}
		return false;
	}
	
	// if theres a conflit it calls wait() on the target
	public synchronized void checkConflict(Point target){
		if(isConflict(target)){
			try { target.wait(); } catch (InterruptedException e) {}
		}
	}
	
	
	// returns minimum distance between this point and another point (x, y)
	private double getDistance(double x, double y){
		
		double a, b, c, d, e, f, g, h, i;
		
		// check all possible distances and return the minimum
		a = getHypoteneuse(x, y-Y_SIZE);
		b = getHypoteneuse(x, y+Y_SIZE);
		c = getHypoteneuse(x, y);
		d = getHypoteneuse(x-X_SIZE, y-Y_SIZE);
		e = getHypoteneuse(x-X_SIZE, y+Y_SIZE);
		f = getHypoteneuse(x-X_SIZE, y);
		g = getHypoteneuse(x+X_SIZE, y-Y_SIZE);
		h = getHypoteneuse(x+X_SIZE, y+Y_SIZE);
		i = getHypoteneuse(x+X_SIZE, y);
		
		return Math.min(Math.min(Math.min(Math.min(Math.min(Math.min(Math.min(Math.min(a, b), c), d), e), f), g), h), i);
	}
	
	// returns the distance between two points on a grid
	private double getHypoteneuse(double x, double y){
		return Math.sqrt(Math.pow(Math.abs(x-xCoordinate),2)+Math.pow(Math.abs(y-yCoordinate),2));
	}
	
	// returns this point's x coordinate
	public double getX() {
		return xCoordinate;
	}
	// returns this point's y coordinate
	public double getY() {
		return yCoordinate;
	}
	
	// if considerZero is not zero, returns random(-1, 0, 1)
	//  otherwise returns random(-1, 1)
	private int getRandomDirection(int considerZero){
		double rand = Math.random();
		if (considerZero!=0){
			if (rand < (1/3)){
				return -1;
			} else if (rand > 2/3){
				return 1;
			} else {
				return 0;
			}
		} else {
			if (rand < 0.5){
				return -1;
			} else {
				return 1;
			}
		}
	}
	
}